#................................................................................
# The MIT License

# Copyright (c) 2009-2010 David Car, david.car7@gmail.com
# Copyright (c) 2009-2010 Michael List, michael.list@gmail.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#................................................................................
from blockit.parsers import pyp as pp
import blockit.basegrammar as gram
import blockit.m4.grammars as m4gram

__doc__ = """
(C) 2010 Michael List, michael.list@gmail.com
(C) 2010 David Car, david.car7@gmail.com

Some of these grammars were hacked from the PyParsing UnderDevelopment site and
likely then follow the same license as pyparsing itself.

"""

# Create forward declarations for possible recursive types
STRUCTDECL = pp.Forward()
ENUMDECL = pp.Forward()

# Basic identifiers and punctuation
CPLUSPLUSLINECOMMENT = pp.Literal("//") + pp.restOfLine
IDENT = pp.Word( pp.alphas+"_", pp.alphanums+"_$" )
INTEGER = pp.Word( pp.nums )
SEMI = pp.Literal(";").suppress()
COMMA = pp.Literal(",").suppress()
LBRACE = pp.Literal("{").suppress()
RBRACE = pp.Literal("}").suppress()
LBRACK = pp.Literal("[").suppress()
RBRACK = pp.Literal("]").suppress()
EQUALS = pp.Literal("=").suppress()
LPAREN = pp.Literal("(").suppress()
RPAREN = pp.Literal(")").suppress()
CONST = pp.Keyword("const")
ELLIPSIS = pp.Literal("...")
ENUM = pp.Keyword("enum")
EXTERN = pp.Keyword("extern").suppress()
LONG = pp.Keyword("long")
SHORT = pp.Keyword("short")
STATIC = pp.Keyword("static").suppress()
STRUCT = pp.Keyword("struct")
VOID = pp.Keyword("void")
ATTRIBUTE = pp.Keyword("__attribute__")
PARENEDVALUE = LPAREN + pp.Word(pp.alphanums) + RPAREN
ATTRIBUTES = pp.Group(ATTRIBUTE + pp.Literal("((") + pp.OneOrMore((pp.Word(pp.alphanums)|PARENEDVALUE)) + pp.Literal("))")).suppress()
INLINE = pp.Keyword("__inline__").suppress()
UNSIGNED = pp.Keyword("unsigned")
POINTERNOTATION = pp.Optional(pp.Word("*"), default=None)
TYPEDEF = pp.Keyword("typedef").suppress()
 
TYPENAME = IDENT
VARNAME = IDENT

ARRAYSIZESPECIFIER = INTEGER | IDENT  
INTRINSICSPEC = pp.Optional(UNSIGNED|CONST) + pp.Optional(pp.OneOrMore(LONG|SHORT|UNSIGNED|CONST)) + \
            pp.oneOf("int long short double char void size_t") + pp.Optional(CONST)
LONGLONGSPEC = pp.Optional(UNSIGNED|CONST) + LONG + LONG # Necessary...
STRUCTTYPESPEC = pp.Optional(CONST) + STRUCT + VARNAME
STRUCTPROTOTYPE = pp.lineStart + STRUCTTYPESPEC + SEMI
ENUMTYPESPEC = pp.Optional(CONST) + ENUM + VARNAME
TYPESPEC = (INTRINSICSPEC|LONGLONGSPEC|STRUCTTYPESPEC|ENUMTYPESPEC|IDENT)

BITFIELDSPEC = ":" + ARRAYSIZESPECIFIER
VARNAMESPEC = ( VARNAME + pp.Optional( BITFIELDSPEC | ( LBRACK + ARRAYSIZESPECIFIER + RBRACK )))
ASSIGNMENTFIELD = pp.Group( VARNAMESPEC + EQUALS + \
    ( gram.HEX_NUM | gram.ENGINEERING_NUM | gram.DECIMAL_NUM | gram.INTEGER | gram.NUM ) )
# For variables with array indices we must have default=None to make sure we put the integers in the right
# spots.  int d=1; int d[1]; both parse to ['d','1'] if you don't. Want int d=1 --> ['d', None, '1']
VARNAMESPECARR = ( VARNAME + pp.Optional( BITFIELDSPEC | ( LBRACK + ARRAYSIZESPECIFIER + RBRACK ), default=None))
ASSIGNMENTFIELDARR = pp.Group( VARNAMESPECARR + EQUALS + \
    ( gram.HEX_NUM | gram.ENGINEERING_NUM | gram.DECIMAL_NUM | gram.INTEGER | gram.NUM ) )


# Specification, i.e. enum cudaSuccess, struct dim3, etc.
ENUMSPEC = ENUM + VARNAMESPEC
STRUCTSPEC = STRUCT + VARNAMESPEC


#.......................................................................
# Structure definitions
#.......................................................................
MEMBERDECL = pp.Group( ( TYPESPEC | TYPENAME | ENUMSPEC ) + POINTERNOTATION + \
    pp.Group( pp.delimitedList( pp.Group(ASSIGNMENTFIELDARR | VARNAMESPECARR) ) ) + SEMI ) | STRUCTDECL
MEMBERENUMDECL = pp.Group( ENUMSPEC + POINTERNOTATION + \
    pp.Group( pp.delimitedList( pp.Group(ASSIGNMENTFIELDARR | VARNAMESPECARR) ) ) + SEMI ) | STRUCTDECL
MEMBERSTRUCTDECL = pp.Group( STRUCTSPEC + POINTERNOTATION + \
    pp.Group( pp.delimitedList( pp.Group(ASSIGNMENTFIELDARR | VARNAMESPECARR) ) ) + SEMI ) | STRUCTDECL
 
# Structs
STRUCTDECL << pp.Group( pp.Optional(TYPEDEF) + STRUCT + pp.Optional(ATTRIBUTES) + pp.Optional(IDENT, default=None) + \
    pp.Group( LBRACE + pp.ZeroOrMore( MEMBERDECL | MEMBERENUMDECL | MEMBERSTRUCTDECL ) + RBRACE) + \
                        POINTERNOTATION + pp.Optional(VARNAMESPEC, default=None) + SEMI )
CSTRUCTBNF = STRUCTDECL
CSTRUCTBNF.ignore( pp.cStyleComment )  # never know where these will crop up!
CSTRUCTBNF.ignore( CPLUSPLUSLINECOMMENT )  # or these either

#.......................................................................
# Enumerator definitions
#.......................................................................
ENUMDECL << pp.Group( pp.Optional(TYPEDEF) + ENUM + pp.Optional(IDENT, default=None) + \
    pp.Group( LBRACE + pp.ZeroOrMore( (ASSIGNMENTFIELD|VARNAMESPEC) + COMMA ) + pp.ZeroOrMore( (ASSIGNMENTFIELD|VARNAMESPEC) ) \
                      + RBRACE ) + POINTERNOTATION + pp.Optional(VARNAMESPEC, default=None) + SEMI )
CENUMBNF = ENUMDECL
CENUMBNF.ignore( pp.cStyleComment )
CENUMBNF.ignore( CPLUSPLUSLINECOMMENT )

#.......................................................................
# Typedef definitions
#.......................................................................
TYPEDEFSPEC = pp.Keyword("typedef") + \
    pp.Group((TYPESPEC ^ STRUCTSPEC ^ ENUMSPEC ^ IDENT) + POINTERNOTATION) + \
    VARNAMESPEC
TYPEDEFDECL = pp.Group( TYPEDEFSPEC + SEMI )
TYPEDEFBNF = TYPEDEFDECL
TYPEDEFBNF.ignore( pp.cStyleComment )
TYPEDEFBNF.ignore( CPLUSPLUSLINECOMMENT )

#.......................................................................
# Function prototypes
#.......................................................................
# The basic prototype is for preprocessed source
PROTOSPEC = pp.Group(pp.Optional(EXTERN|STATIC) + TYPESPEC + POINTERNOTATION ) + \
    VARNAME + pp.Group( LPAREN + \
    (pp.delimitedList(pp.Group(TYPESPEC + POINTERNOTATION + VARNAMESPEC)|ELLIPSIS, delim=',')|VOID) + \
    RPAREN )
# The generic prototype is for unprocessed source
GENERICPROTOSPEC = pp.OneOrMore((EXTERN | TYPESPEC | VARNAME) + POINTERNOTATION) + pp.Group( LPAREN + \
    (pp.delimitedList(pp.Group(TYPESPEC + POINTERNOTATION + VARNAMESPEC)|ELLIPSIS, delim=',')|VOID) + \
    RPAREN )

#PROTODECL = pp.LineStart() + pp.Group( PROTOSPEC + SEMI )
PROTODECL = pp.Group( PROTOSPEC + SEMI )
PROTOBNF = PROTODECL
PROTOBNF.ignore( pp.cStyleComment )
PROTOBNF.ignore( CPLUSPLUSLINECOMMENT )
#GPROTODECL = pp.LineStart() + pp.Group( GENERICPROTOSPEC + SEMI )
GPROTODECL = pp.Group( GENERICPROTOSPEC + SEMI )
GPROTOBNF = GPROTODECL
GPROTOBNF.ignore( pp.cStyleComment )
GPROTOBNF.ignore( CPLUSPLUSLINECOMMENT )
 
if __name__ == "__main__":

    import unittest

#=======================================================================
    class TestCGrammar(unittest.TestCase):
#=======================================================================
        testData1 = """
        struct {
            long a;
            short b;
            char c[32]; /* comment */
            } a;

        struct {
          int i;
          } dimple;
    """
 
        testData2 = """
        union {
            long a;
            long* a;
            long** a;
            long*** a;
            long**** a;
            long *a;
            long **a;
            struct {
                int x;
                int y;
                } pt;  // this is an embedded struct
            struct {
                int x,y;
                struct {
                   char* a;
                   char* b;
                   } inner;
                } pt2;
            struct {
                int x;
                int y;
                }* coordPtr; /* this is just a pointer to a struct */
            short b;
            char c[32];
            char d[MAX_LENGTH /* + 1 to make room for terminating null */ ];
            char* name;
            char *name2;  /* no one can agree where the '*' should go */
            int bitfield:5;  /* this is rare, but not hard to add to parse grammar */
            int bitfield2:BIT2LEN;
            void* otherData;
            } a;
    """
 
        testData3 = """
    struct cudaChannelFormatDesc
    {
      int                        x; /**< x */
      int                        y; /**< y */
      int                        z; /**< z */
      int                        w; /**< w */
      enum cudaChannelFormatKind f; /**< Channel format kind */
    };

    struct cudaArray;

    struct cudaPitchedPtr
    {
      void   *ptr;      /**< Pointer to allocated memory */
      size_t  pitch;    /**< Pitch of allocated memory in bytes */
      size_t  xsize;    /**< Logical width of allocation in elements */
      size_t  ysize;    /**< Logical height of allocation in elements */
    };

    struct cudaExtent
    {
      size_t width;     /**< Width in elements when referring to array memory, in bytes when referring to linear memory */
      size_t height;    /**< Height in elements */
      size_t depth;     /**< Depth in elements */
    };

    struct cudaPos
    {
      size_t x;     /**< x */
      size_t y;     /**< y */
      size_t z;     /**< z */
    };

    struct cudaMemcpy3DParms
    {
      struct cudaArray      *srcArray;  /**< Source memory address */
      struct cudaPos         srcPos;    /**< Source position offset */
      struct cudaPitchedPtr  srcPtr;    /**< Pitched source memory address */

      struct cudaArray      *dstArray;  /**< Destination memory address */
      struct cudaPos         dstPos;    /**< Destination position offset */
      struct cudaPitchedPtr  dstPtr;    /**< Pitched destination memory address */

      struct cudaExtent      extent;    /**< Requested memory copy size */
      enum cudaMemcpyKind    kind;      /**< Type of transfer */
    };
    struct cudaFuncAttributes
    {
       size_t sharedSizeBytes;
       size_t constSizeBytes;
       size_t localSizeBytes;
       int maxThreadsPerBlock;
       int numRegs;
       int ptxVersion;
       int binaryVersion;
       int __cudaReserved[6];
    };

    struct cudaDeviceProp
    {
      char   name[256];                 /**< ASCII string identifying device */
      size_t totalGlobalMem;            /**< Global memory available on device in bytes */
      size_t sharedMemPerBlock;         /**< Shared memory available per block in bytes */
      int    regsPerBlock;              /**< 32-bit registers available per block */
      int    warpSize;                  /**< Warp size in threads */
      size_t memPitch;                  /**< Maximum pitch in bytes allowed by memory copies */
      int    maxThreadsPerBlock;        /**< Maximum number of threads per block */
      int    maxThreadsDim[3];          /**< Maximum size of each dimension of a block */
      int    maxGridSize[3];            /**< Maximum size of each dimension of a grid */
      int    clockRate;                 /**< Clock frequency in kilohertz */
      size_t totalConstMem;             /**< Constant memory available on device in bytes */
      int    major;                     /**< Major compute capability */
      int    minor;                     /**< Minor compute capability */
      size_t textureAlignment;          /**< Alignment requirement for textures */
      int    deviceOverlap;             /**< Device can concurrently copy memory and execute a kernel */
      int    multiProcessorCount;       /**< Number of multiprocessors on device */
      int    kernelExecTimeoutEnabled;  /**< Specified whether there is a run time limit on kernels */
      int    integrated;                /**< Device is integrated as opposed to discrete */
      int    canMapHostMemory;          /**< Device can map host memory with cudaHostAlloc/cudaHostGetDevicePointer */
      int    computeMode;               /**< Compute mode (See ::cudaComputeMode) */
      int    maxTexture1D;              /**< Maximum 1D texture size */
      int    maxTexture2D[2];           /**< Maximum 2D texture dimensions */
      int    maxTexture3D[3];           /**< Maximum 3D texture dimensions */
      int    maxTexture2DArray[3];      /**< Maximum 2D texture array dimensions */
      size_t surfaceAlignment;          /**< Alignment requirements for surfaces */
      int    concurrentKernels;         /**< Device can possibly execute multiple kernels concurrently */
      int    ECCEnabled;                /**< Device has ECC support enabled */
      int    pciBusID;                  /**< PCI bus ID of the device */
      int    pciDeviceID;               /**< PCI device ID of the device */
      int    tccDriver;                 /**< 1 if device is a Tesla device using TCC driver, 0 otherwise */
      int    __cudaReserved[21];
    };

    struct;

    typedef struct aTypeDef ThisIsWRONG;
    """
    
        def setUp(self): pass

        def tearDown(self): pass

    #-------------------------------------------------------------------
        def testScanMultiplePrototype(self):
    #-------------------------------------------------------------------
            """Checks scanning of multiple preprocessed prototypes"""

            bnf = PROTOBNF
            ctr = 0
            for m,s,e in bnf.scanString("""
                int cg_state_read(char **StateDescription);
                int cg_state_write(char const * StateDescription);
                """):
                ctr += 1
            self.failUnlessEqual(ctr, 2, "Failed to find correct number of matches")


    #-------------------------------------------------------------------
        def testParsePrototype(self):
    #-------------------------------------------------------------------
            """Checks parsing of preprocessed prototypes"""

            bnf = PROTOBNF
            try:
                output = bnf.parseString("extern cudaError_t cudaGetDeviceProperties(struct cudaDeviceProp *prop, int device);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("CUresult cuGraphicsMapResources(unsigned int count, CUgraphicsResource *resources, CUstream hStream);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("int myfunc(struct jdog *damoney, ...);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("extern cudaError_t \
                    cudaGraphicsSubResourceGetMappedArray(struct cudaArray **array, \
                    cudaGraphicsResource_t resource, unsigned int arrayIndex, unsigned int mipLevel);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("int cg_state_read(char **StateDescription);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("int cg_state_write(char const * StateDescription);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),3,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0][0]),2,"Output dimensions incorrect")
            self.failUnlessEqual(output[0][1],"cg_state_write","Incorrect prototype name found!")


    #-------------------------------------------------------------------
        def testParseGenericPrototype(self):
    #-------------------------------------------------------------------
            """Checks parsing of non-preprocessed prototypes"""

            bnf = GPROTOBNF
            try:
                output = bnf.parseString("extern __host__ cudaError_t CUDARTAPI cudaGetDevice(void *devPtr);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("extern __host__ cudaError_t CUDARTAPI cudaSetDevice(int *device);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("extern __host__ cudaError_t __stdcall \
                    cudaHostGetDevicePointer(void **pDevice, void *pHost, unsigned int flags);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("extern cudaError_t cudaGetDeviceProperties(struct cudaDeviceProp *prop, int device);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("CUresult cuGraphicsMapResources(unsigned int count, CUgraphicsResource *resources, CUstream hStream);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("extern cudaError_t \
                    cudaGraphicsSubResourceGetMappedArray(struct cudaArray **array, \
                    cudaGraphicsResource_t resource, unsigned int arrayIndex, unsigned int mipLevel);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("int cg_state_read(char **StateDescription);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("int cg_state_write(char const * StateDescription);")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")


    #-------------------------------------------------------------------
        def testParseStructTypedef(self):
    #-------------------------------------------------------------------
            """Checks parsing of struct defined in typedef"""

            bnf = CSTRUCTBNF
            try:
                output = bnf.parseString("""
typedef struct cuda_error {
 int success; /** something */
 long failure; /** something */
 unsigned short int badness; /** something */
} CUresult;
                """)
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")


    #-------------------------------------------------------------------
        def testParseEnumTypedef(self):
    #-------------------------------------------------------------------
            """Checks parsing of enum defined in typedef"""

            bnf = CENUMBNF

            try:
                output = bnf.parseString("""
typedef enum cuda_error {
 success = 1, /** something */
 failure = 2, /** something */
 badness = 3 /** something */
} CUresult;
                """)
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")

            try:
                output = bnf.parseString("""
typedef enum { success = 1, failure = 2, badness = 3 } CUresult_t;
                """)
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")


    #-------------------------------------------------------------------
        def testParseBasicTypedef(self):
    #-------------------------------------------------------------------
            """Checks to ensure proper parsing of simple typedefs"""

            bnf = TYPEDEFBNF

            try:
                output = bnf.parseString("typedef int int_t;")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),3,"Length of result inappropriate")

            try:
                output = bnf.parseString("typedef void* GLPtr;")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),3,"Length of result inappropriate")

            try:
                output = bnf.parseString("typedef struct cudaError cudaError_t;")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),3,"Length of result inappropriate")

            try:
                output = bnf.parseString("typedef enum myEnum cuda_returns;")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),3,"Length of result inappropriate")

            try:
                output = bnf.parseString("typedef unsigned long long CUdeviceptr;")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),3,"Length of result inappropriate")

            try:
                output = bnf.parseString("typedef double2 cuDoubleComplex;")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),3,"Length of result inappropriate")
            self.failUnlessEqual(output[0][0], "typedef")
            self.failUnlessEqual(len(output[0][1]), 2)
            self.failUnlessEqual(output[0][2], "cuDoubleComplex")


    #-------------------------------------------------------------------
        def testParseBasicEnum(self):
    #-------------------------------------------------------------------
            """Checks to ensure proper parsing of simple enums"""

            bnf = CENUMBNF

            try:
                output = bnf.parseString("enum { my = 3, yours = 0x01 };")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),5,"Length of result inappropriate")

            try:
                output = bnf.parseString("enum dum { a, b, c, d };")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),5,"Length of result inappropriate")


    #-------------------------------------------------------------------
        def testParseBasicStruct(self):
    #-------------------------------------------------------------------
            """Checks to ensure proper parsing of simple structs"""

            bnf = CSTRUCTBNF
            bnf.ignore( m4gram.POUNDLINE )
            try:
                output = bnf.parseString("struct {\nunsigned int x, y, z;\n\
                    \n __host__ something something;\n __host__ something something;\n };")
                self.fail("Parsed invalid string")
            except:
                pass

            try:
                output = bnf.parseString("struct my_struct {\nunsigned int x, y, z;\n\
                    \n something something;\n something something;\n };")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),5,"Length of result inappropriate")

            try:
                output = bnf.parseString("struct my_struct {\nunsigned int x, y, z;\n\
                    \n something something[15];\n something something=4;\n };")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Output dimensions incorrect")
            self.failUnlessEqual(len(output[0]),5,"Length of result inappropriate")

    #-------------------------------------------------------------------
        def testParseStruct(self):
    #-------------------------------------------------------------------
            """Checks parsing of structs"""

            bnf = CSTRUCTBNF

            ctr = 0
            for match,s,e in bnf.scanString( self.testData1 ):
                ctr += 1
                self.failUnlessEqual(len(match),1,"Match structure is too long")
                self.assertTrue(len(match[0]) == 5,"Match result not correct")
            self.failUnlessEqual(ctr,2,"Not enough structs parsed")

            ctr = 0
            for match,s,e in bnf.scanString( self.testData2 ):
                ctr += 1
                self.failUnlessEqual(len(match),1,"Match structure is too long")
                self.assertTrue(len(match[0]) == 5,"Match result not correct")
            self.failUnlessEqual(ctr,3,"Wrong number of structs parsed")

            ctr = 0
            for match,s,e in bnf.scanString( self.testData3 ):
                ctr += 1
                self.failUnlessEqual(len(match),1,"Match structure is too long")
                self.assertTrue(len(match[0]) == 5,"Match result not correct")
            self.failUnlessEqual(ctr,7,"Not enough structs parsed")

            try:
                output = bnf.parseString("struct mystruct {\nchar name[256];\nint i;\ndouble d=4.0, f=3.0;\n};")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output[0]),5,"Match result not correct")

            try:
                output = bnf.parseString("struct __attribute__((aligned (8))) mystruct {};")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output[0]),5,"Match result not correct")


    #-------------------------------------------------------------------
        def testParseStructPrototype(self):
    #-------------------------------------------------------------------
            """Checks parsing of struct prototypes"""

            bnf = STRUCTPROTOTYPE

            try:
                output = bnf.parseString( "struct mystruct;" )
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),2,"Match structure is too long")
            self.failUnlessEqual(output[0],"struct","Failed to parse correctly")
            self.failUnlessEqual(output[1],"mystruct","Failed to parse correctly")


    #-------------------------------------------------------------------
        def testVarNameSpec(self):
    #-------------------------------------------------------------------
            """Checks parsing of variable names"""

            bnf = VARNAMESPEC

            try:
                output = bnf.parseString("myvar[16]")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),2,"Did not parse string")

            try:
                output = bnf.parseString("myvar")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),1,"Did not parse string properly")


    #-------------------------------------------------------------------
        def testIntrinsicSpec(self):
    #-------------------------------------------------------------------
            """Tests capture of intrinsic variable declaration"""

            bnf = INTRINSICSPEC

            try:
                output = bnf.parseString("unsigned long int")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),3)
            self.failUnlessEqual(output[0],"unsigned")
            self.failUnlessEqual(output[1],"long")
            self.failUnlessEqual(output[2],"int")

            try:
                output = bnf.parseString("unsigned long long int")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),4)
            self.failUnlessEqual(output[0],"unsigned")
            self.failUnlessEqual(output[1],"long")
            self.failUnlessEqual(output[2],"long")
            self.failUnlessEqual(output[3],"int")


    #-------------------------------------------------------------------
        def testTypeSpec(self):
    #-------------------------------------------------------------------
            """Tests type specification capturing"""

            bnf = TYPESPEC

            try:
                output = bnf.parseString("unsigned long long")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),3)
            self.failUnlessEqual(output[0],"unsigned")
            self.failUnlessEqual(output[1],"long")
            self.failUnlessEqual(output[2],"long")

            try:
                output = bnf.parseString("char const")
            except:
                self.fail("Failed to parse string")
            self.failUnlessEqual(len(output),2)
            self.failUnlessEqual(output[0],"char")
            self.failUnlessEqual(output[1],"const")


    suite = unittest.TestLoader().loadTestsFromTestCase(TestCGrammar)
    unittest.TextTestRunner(verbosity=2).run(suite)
