#............................................................................
# Copyright (c) 2009,2010,2011 David Car, david.car7@gmail.com
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#..........................................................................
__all__ = ['propagateTemplateDeps',
           'realizeTemplates',
           'buildSTL']

from blockit.tables import SymbolTable
from blockit.funcs import *
from blockit.parsers import pyp
from blockit.sets import ASet
from blocks import *

import os

#................................................................................
# Global functions and stuff
#................................................................................
# Built-in type list
header='''!...............................................................................
!.
!. This code was pre-parsed by PyF95++, a block parser and code generator.
!.
!................................................................................'''
header += os.linesep

#................................................................................
def buildSTL( fileTable, globalTable, destDir=None ):
    '''Build a Standard Template Library from all the templated modules within
    the fileTable.  This is necessary because you can run into circular
    dependencies between files based on template instantiations.  Therefore,
    you either need to create a separate file for each instantiation or you
    can put them within a single file and order them appropriately.  The
    latter is the current method.

    fileTable : symbol table containing all the files
    globalTable: global symbol table
    breakOut: boolean on whether to create multiple files or a single STL file [False]

    '''
#     import profile
    table = SymbolTable()
    table.registerAll( fileTable )
    tempMods = table.filter(And(And(isModule, hasTempParent),
                                Not(isTemplated)))

    #................................................................................
    # If the templated modules parent came from the cache, then set the
    # fromCache property on the modules
    #................................................................................
    for _mod in tempMods.values():
        if _mod.parent().fromCache:
            _mod.fromCache = True

    tempModParents = list(set([x.parent() for x in tempMods.values()]))
    if not tempModParents:
        return None

    #................................................................................
    # Generate a depGraph between the template modules and a build order
    #................................................................................
    changedBlocks = set()
    for blk in tempMods.values():
        blkFile = blk.upTo( AFile )
        changedBlocks.add( blk.parent() )
        changedBlocks.add( blkFile )
        prefix, ext = os.path.splitext( blkFile.fileName() )
        srcDir = os.path.split(prefix)[0]
        # tmp = AFile( [ blk.name() + ext, blk.hash() ] )
        # Hash is no longer used so set to zero here
        tmp = AFile( [ blk.name() + ext, '0' ] )
        tmp.setNameLength( blk.getNameLength() )
        print 'Creating new AFile: %s' % ( tmp.fileName(), )
        tmp.addChild( blk.parent(), regen=False )
        # profile.runctx( "tmp.addChild( blk, regen=False )", globals(), locals() )
        tmp.setFileName( os.path.join( srcDir, tmp.name() ) )
        tmp.fromCache = blk.fromCache
        # tmp.directives = blk.directives
        fileTable.register( tmp )
        
    for blk in changedBlocks:
        print 'Regenerating child deps for block: (%s)' % ( blk.name(), )
        blk.regenChildDeps()
        
#................................................................................
def propagateTemplateDeps( nonTempBlocks, tempBlocks, IgnoreSymbols=ASet() ):
    '''Propagate template parameters from non-templated blocks that instantiate a
    templated type to the parent templated block, i.e. the block at the root
    of a nested sequence of blocks (e.g. a templated module will have all
    blocks under it templated).
    
    nonTempBlocks : symbol table of blocks that are not templated and
                    have template dependencies

    tempBlocks : complete (all blocks expanded) symbol table of blocks that are
                 templated
    IgnoreSymbols: a set of symbols that should not be propagated and instantiated.

    '''
    #................................................................................
    global globalSymbolMap

    joinChar = tempBlocks.joinCharacter()
    for k,v in nonTempBlocks.items():
        dep = v.dependencies()

        for typ in dep:
            #..................................................................
            # Order of operations:
            # 1. Check if user specified external symbols in a directive
            # that may have a wildcard.
            # 2. Resolve any wildcards in the symbol.
            # 3. Parse if its a template spec.
            # 4. Add the resolved name to the global symbol map.
            # 5. Check if the resolved symbol is an external symbol.
            #..................................................................
            if typ in IgnoreSymbols: continue
            _typ = v.resolveSymbol( typ, mangle=False )
            pairs = parseTemplateSpecString( _typ )
            typ = _typ
            if typ in IgnoreSymbols: continue

            while len( pairs ) > 0:
                name, spec = pairs.pop()
                tmp = tempBlocks.filter(And(Or(Or(isAType, isFunction),isSubroutine), 
                                            startsWith(name+joinChar))).values()

                if len( tmp ) == 0:
                    msg = "Missing dependency: %s<%s>%s" % ( name, spec, os.linesep )
                    msg += "For block %s%s" % ( v.name(), os.linesep )
                    continue
                    # raise Exception( msg )

                blk = len( tmp ) > 1 and \
                      templateMRO( v, tmp, spec, v.getSig( typ ) ) or \
                      tmp[ 0 ]

                spec = templateSpecToTuple( spec )
                blk.addObjects( spec )
                spec = ','.join( [ x for x in spec if x not in
                                   IgnoreSymbols ] )
                pairs.extend( parseTemplateSpecString( spec ) )

#................................................................................
def templateMRO( caller, tmp, spec, sig=None ):
    '''Given a list of blocks, a template spec and a optional
    function/subroutine signature, find the block that matches.

    '''
    # Filters for resolving ambiguous dependencies
    # Number of template specifications
    def specLength(n): return lambda v: len(v.template()) == n

    # If parent is template
    zeroDepth = lambda v: isinstance(v.parent(), Template)
    
    # Depth of nesting
    def _depth(typ, c):
        if not f2(typ):
            c += 1
            depth(t, c)
        return c

    def depth(typ):
        return _depth( typ, 0 )

    #....................................................................
    # More than one possibility, so try to resolve with mro list
    # 1. Check that the template length is the same.
    # 2. Check if the parent is the template.
    #....................................................................
    name = tmp[0].name()
    mro = ( specLength( len( templateSpecToTuple( spec ) ) ), zeroDepth )
    for test in mro:
        tmp = [ x for x in tmp if test( x ) ]
        if len( tmp ) == 1: return tmp[ 0 ]

    #....................................................................
    # Check the function/subroutine type signature for a match
    #....................................................................
    if sig:
        for x in tmp:
            _args = x.argList
            if not ( len( _args ) == len( sig ) ): continue
            _syms = x.symbols
            _args = [ _syms[ y ] for y in _args ]
            if all( map( lambda x, y: x == y, _args, sig  ) ): return x

    #......................................................................
    # Raise exception, no match
    #......................................................................
    if ( len( tmp ) > 1 ):
        print "Ambiguous dependencies found:"
        for t in tmp:
             print "(%s) in (%s)" % ( t.name(), t.upTo( AFile ).name() )
        raise Exception()
    elif ( len( tmp ) == 0 ):
        msg = "Missing dependency: %s<%s>%s" % ( name, spec, os.linesep )
        msg += "For block %s%s" % ( caller.name(), os.linesep )
        raise Exception( msg )

#................................................................................
def realizeTemplates( table, cache, IgnoreSymbols=ASet() ):
    '''Find all the root templated blocks and loop through them:
    
    table : global symbol table
    lib : FortranLibrary CacheManager object

    Logic
    -----
    1. Resolve their objects (i.e. any wildcards and such).

    2. Begin main loop and propagate these objects to their children and
    discover any new types that are created from the templates internally.

    3. Propagate these new types to their corresponding template blocks
    objects sets.

    4. Goto to 2 and repeat until no object sets of any template have changed
    in any of the templated types.

    Returns a new SymbolTable() with the instantiated templates.

    '''
    
    tempTable = SymbolTable()
    realized = SymbolTable()

    # Root template blocks
    rootTempBlocks = table.filter( isTemplate )

    # import pdb; pdb.set_trace()
    # All leaf templated blocks
    # leafTempBlocks = table.filter(And(isTemplated, isLeaf))
    leafTempBlocks = table.filter( And( isTemplated, Not( isTemplate ) ) )
    
    #................................................................................
    # Recursion termination: When objects sets of all root templated blocks
    # stops changing.
    #................................................................................
    anyChanged = True
    while anyChanged:
        anyChanged = False

        for sym, blk in rootTempBlocks.items():
            objSet = blk.objects()
            if not objSet.changed(): continue
            _objSet = objSet.changeSet()
            blk.objects().reset()
            for params in _objSet:
                print 'From blk: %s'%( blk.name(), )
                print 'Instantiating (%s) with parameters %s'% \
                    (','.join( blk.children() ), str( params ) )
                _blk = blk.instantiate( params, cache )
                
                #.............................................................
                # Temporarily add _blk to the realized table and call
                # propagateTemplateDeps to push it's dependencies to the other
                # blocks.  Then add it to the returned tempTable symbol table.
                #.............................................................
                realized.register( _blk )
                propagateTemplateDeps( realized, 
                                       leafTempBlocks, 
                                       IgnoreSymbols )
                tempTable.register( realized.pop( _blk.name() ) )

            anyChanged = True

    return tempTable


