#............................................................................
# Copyright (c) 2009,2010,2011 David Car, david.car7@gmail.com
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA 02111-1307 USA
#..........................................................................
__all__ = ['FortranLibrary',
           'FortranCacheManager',
           'getAllLibKeys',
           'libCheck']

from blockit.library import Library, CacheManager
from blocks import *
from blockit.tables import *

import os
import sys
import time
import glob
import pickle

#................................................................................
# Helper functions
#................................................................................
def getAllLibKeys(lib):
    '''Returns a triple of the sets of library keys: (file names, template
    lists, template instantiations).  Strips the 'METADATA' key.

    '''
    _keys = lib.keys()
    _keys = set(_keys)
    files = set([key for key in _keys if key.lower().endswith('.f90t') or 
                 key.lower().endswith('.f90') or 
                 key.lower().endswith('.inc')])
    tempLists = set([key for key in _keys if key.endswith('/templates')])
    tables = set( [ key for key in _keys 
                      if key in ( 'HASHTABLE', 'SYMBOLTABLE' ) ] )
    
    templates = _keys
    for x in (files, tempLists, tables):
        templates = templates.difference( x )

#python2.6 only    templates = _keys.difference(files, tempLists, tables)
    templates.discard('METADATA')

    return (files, tempLists, templates, tables)

def libCheck(lib):
    '''Verify that the library has all instantiated templates are associated
    with a file and that all files exist.

    '''
    failed = False
    files, tempLists, templates, tables = getAllLibKeys(lib)

    print '.'*80
    print 'Verifying Library Integrity'.center(80)
    print '.'*80

    print 'Found %d files'%(len(files),)
    print 'Found %d template lists'%(len(tempLists),)
    print 'Found %d template instantiations'%(len(templates),)

    for f in files:
        try:
            _listing = lib[f+'/templates']
            tempLists.discard(f+'/templates')
            for temp in _listing:
                try:
                    templates.discard(temp)
                except:
                    print '''%s is missing from database for file
            %s'''%(temp[:10]+'...'+temp[temp.rfind('/'):], f)
        except:
            pass

    if len(tempLists) > 0:
        failed = True
        print "Found %d orphaned template listings"%(len(tempLists),)
        for temp in tempLists:
            print '-> %s is orphaned'%(temp,)

    if len(templates) > 0:
        failed = True
        print "Found %d orphaned templates"%(len(templates),)
        for temp in templates:
            print '-> %s is orphaned'%(temp[:20]+'...'+temp[temp.rfind('/'):],)
    
    if not failed:
        print 'No orphan template listings found'
        print 'No orphan templates found'
        print 'Passed verification'
    else:
        print 'Verification failed'
    
    print '.'*80
    return (tempLists, templates)

#................................................................................
class FortranLibrary(Library):
    '''Extends blockit.library.Library class to include template storage.

    '''

    VERSION = '1.4'
    def __init__(self, name, blockList):
        super(FortranLibrary, self).__init__(name, blockList)

    def getFile( self, name, hsh, checkHash=True ):
        '''Gets a File block from the database.

        name : the file name
        hsh : the hash of the file to check against the one stored in the
              library.

        '''
        _blk = None
        _hsh, _tree = self._lib.get( name, ( None, None ) )
        if _tree is not None:
            #......................................................................
            # Sanity check to make sure the hash matches with the key name
            # hash (only if checkHash is True)
            #......................................................................
            if checkHash:
                _blk = ( hsh == _hsh ) and self.unpackTree( _tree ) or None
            else:
                _blk = self.unpackTree( _tree )

            if _blk:
                _blk = ( _blk.version() == FortranBlock.VERSION ) and _blk or None

        return _blk

    def storeFile(self, blk):
        '''Saves a AFile block to the database.

        '''
        #.........................................................................
        # Check if the file has instaniated templates.  If so, must remove
        # them for the library since its being restored and therefore most
        # likely has changed
        #.........................................................................
        _tempList = self._lib.get(blk.sourceFileName()+'/templates', [])
        for _temp in _tempList:
            try:
                self._lib.pop(_temp)
            except:
                pass
            
        if _tempList: self._lib.pop(blk.sourceFileName()+'/templates')
        self._lib[blk.sourceFileName()] = (blk.hash(), blk.getDictTree())

    def getTemplate(self, name):
        '''Get an instantiated template from the library.

        '''
        _tree = self._lib.get(name, None)
        if _tree:
            return self.unpackTree(_tree)
        return _tree

    def storeTemplate(self, blk):
        '''Store an instantiated template in the library'''

        #......................................................................
        # Get the AFile blk that encapsulates this template
        #......................................................................
        blk.fromLib = False
        _parentFile = blk.upTo(AFile)
        if _parentFile is None:
            raise Exception('''Template is not contained within a file.  Cannot
        generate library key''')

        _ky = _parentFile.sourceFileName()+'/templates'
        _tempList = self._lib.get(_ky, None)
        if _tempList is None:
            self._lib[_ky] = [blk.hash()]
        elif blk.hash() not in _tempList:
            _tempList.append(blk.hash())
            self._lib[_ky] = _tempList

        self._lib[blk.hash()] = blk.getDictTree()

    def storeHashTable(self, table):
        '''Store dictionary of filenames and their hash values.

        '''
        if not isinstance( table, SymbolTable ):
            _table = SymbolTable()
            _table.update(table)
            self._lib[ 'HASHTABLE' ] = _table
        else:
            self._lib[ 'HASHTABLE' ] = table

    def getHashTable(self):
        '''Retrieve the file hash table from the library.

        '''
        return self._lib.get( 'HASHTABLE', SymbolTable() )

    def storeSymbolTable(self, table):
        '''Store dictionary of all the symbols.

        '''
        if not isinstance( table, SymbolTable ):
            _table = SymbolTable()
            _table.update(table)
            self._lib[ 'SYMBOLTABLE' ] = _table
        else:
            self._lib[ 'SYMBOLTABLE' ] = table

    def getSymbolTable(self):
        '''Retrieve the symbol table from the library.

        '''
        return self._lib.get( 'SYMBOLTABLE', SymbolTable() )

    def _upgradeLib( self ):
        '''Function called when a library version number is less than the the
        current codes version number for the library.

        '''
        print '''MESSAGE: Library needs to be rebuilt due to new metadata for
        blocks and the library itself.  Sorry for the inconvenience.'''
        if self.VERSION < '1.4': self._lib.clear()

    def __iadd__(self, other):
        '''Intelligently merge two FortranLibrary instances.  

        '''
        raise NotImplementedError("Stub:  Need to finish this method")

    @staticmethod
    def genKey(blk):
        '''Static method to generate an instantiated Template block key for
        the library.  The key is generated from the parent file's hash as
        follows: 

        <parent file hash value>/templates

        Input
        -----
        blk : an instantiated template block

        '''
        
        if not isinstance(blk, Template):
            return Library.genKey(blk)

        _parentFile = blk.upTo(AFile)
        if _parentFile is None:
            raise Exception('''Template is not contained within a file.  Cannot
        generate library key''')

        return _parentFile.hash()+'/templates'

#================================================================================
class FortranCacheManager( CacheManager ):

    def genCacheName( self, blk ):
        '''Generate the cache file name from a block.

        '''
        parentFile = blk.upTo( AFile )
        parentFileName = os.path.abspath( parentFile.sourceFileName() )
        parentBaseName = os.path.basename( parentFileName )
        parentDir = os.path.dirname( parentFileName )
        cachedDir = os.path.join( parentDir, ( '.' + parentBaseName ) )
        if not os.path.exists( cachedDir ):
            try:
                os.mkdir( cachedDir )
            except:
                e = sys.exc_info()[1]
                sys.stderr.write(
'''Cannot find directory for %s%s''' % ( parentFileName, os.linesep ) )
                sys.stderr.write('Error:  %s%s''' % ( e, 
                                                      os.linesep ) )

        cacheName = os.path.join( cachedDir, blk.name() + self._ext ) 
        return ( self.checkStat( parentFileName, 
                                 cacheName ),
                 cacheName )

    def checkStat( self, parent, cache ):
        '''Check if the cache is newer than its source file and return
        boolean.

        '''
        if os.path.exists( cache ):
            fileInfo = os.stat( parent )
            cacheInfo = os.stat( cache)
            return ( fileInfo.st_ctime < cacheInfo.st_ctime, True )
        return ( False, False )
